from .base import RecordEncoder
from .cohere import CohereRecordEncoder
from .dense import DenseRecordEncoder
from .openai import OpenAIRecordEncoder
from .anyscale import AnyscaleRecordEncoder
from .azure_openai import AzureOpenAIRecordEncoder
from .jina import JinaRecordEncoder
from .sentence_transformers import SentenceTransformerRecordEncoder
from .hybrid import HybridRecordEncoder



from typing import List, Optional, cast

from canopy.chat_engine.query_generator import QueryGenerator
from canopy.chat_engine.history_pruner.raising import RaisingHistoryPruner
from canopy.llm import BaseLLM, CohereLLM
from canopy.models.data_models import Messages, Query


class CohereQueryGenerator(QueryGenerator):
    """
    Query generator for LLM clients that have a built-in feature to
    generate search queries from chat messages.
    """
    _DEFAULT_COMPONENTS = {
        "llm": CohereLLM,
    }

    def __init__(self,
                 *,
                 llm: Optional[BaseLLM] = None):
        self._llm = llm or self._DEFAULT_COMPONENTS["llm"]()

        if not isinstance(self._llm, CohereLLM):
            raise NotImplementedError(
                "CohereQueryGenerator only compatible with CohereLLM"
            )

        self._history_pruner = RaisingHistoryPruner()

    def generate(self,
                 messages: Messages,
                 max_prompt_tokens: int) -> List[Query]:
        messages = self._history_pruner.build(chat_history=messages,
                                              max_tokens=max_prompt_tokens)
        llm = cast(CohereLLM, self._llm)
        queries = llm.generate_search_queries(messages)
        return [Query(text=query) for query in queries]

    async def agenerate(self,
                        messages: Messages,
                        max_prompt_tokens: int) -> List[Query]:
        raise NotImplementedError



from .base import QueryGenerator
from .function_calling import FunctionCallingQueryGenerator
from .last_message import LastMessageQueryGenerator
from .instruction import InstructionQueryGenerator
from .cohere import CohereQueryGenerator



from typing import Optional

from canopy.chat_engine.history_pruner.base import HistoryPruner
from canopy.models.data_models import Messages, Context


class RaisingHistoryPruner(HistoryPruner):

    def build(self,
              chat_history: Messages,
              max_tokens: int,
              system_prompt: Optional[str] = None,
              context: Optional[Context] = None, ) -> Messages:
        max_tokens = self._max_tokens_history(max_tokens,
                                              system_prompt,
                                              context)
        token_count = self._tokenizer.messages_token_count(chat_history)
        if token_count > max_tokens:
            raise ValueError(f"The history require {token_count} tokens, "
                             f"which exceeds the calculated limit for history "
                             f"of {max_tokens} tokens left for"
                             f" history out of {max_tokens} tokens"
                             f" allowed in context window.")
        return chat_history

    async def abuild(self,
                     chat_history: Messages,
                     max_tokens: int) -> Messages:
        raise NotImplementedError()



import time
from copy import deepcopy
from typing import Union, Iterable, Optional, Any, Dict, List

from tenacity import retry, stop_after_attempt

try:
    import cohere
except (OSError, ImportError, ModuleNotFoundError):
    _cohere_installed = False
else:
    _cohere_installed = True

from canopy.llm import BaseLLM
from canopy.llm.models import Function
from canopy.models.api_models import (
    _Choice,
    _StreamChoice,
    ChatResponse,
    StreamingChatChunk,
    TokenCounts,
)
from canopy.models.data_models import Context, MessageBase, Messages, Role, Query
from canopy.context_engine.context_builder.stuffing import StuffingContextContent


COMMON_PARAMS = {
    "model",
    "frequency_penalty",
    "logit_bias",
    "max_tokens",
    "presence_penalty",
    "stream",
    "temperature",
}


EQUIVALENT_PARAMS = {
    "top_p": "p",
    "user": "user_name",
}


class CohereLLM(BaseLLM):
    """
    Cohere LLM wrapper built on top of the Cohere Python client.

    Note: Cohere requires a valid API key to use this class.
          You can set the "CO_API_KEY" environment variable to your API key.
    """
    def __init__(self,
                 model_name: str = "command",
                 *,
                 api_key: Optional[str] = None,
                 base_url: Optional[str] = None,
                 ignore_unrecognized_params: Optional[bool] = False,
                 **kwargs: Any,
                 ):
        """
        Initialize the Cohere LLM.

        Args:
            model_name: The name of the model to use. See https://docs.cohere.com/docs/models
            api_key: Your Cohere API key. Defaults to None (uses the "CO_API_KEY" environment variable).
            base_url: The base URL to use for the Cohere API. Defaults to None (uses the "CO_API_URL" environment variable if set, otherwise use default Cohere API URL).
            ignore_unrecognized_params: Flag to suppress errors when unrecognized model params (from other LLMs) are passed to Cohere.
            **kwargs: Generation default parameters to use for each request. See https://platform.openai.com/docs/api-reference/chat/create
                    For example, you can set the temperature, p, etc
                    These params can be overridden by passing a `model_params` argument to the `chat_completion` methods.
        """  # noqa: E501
        super().__init__(model_name)

        if not _cohere_installed:
            raise ImportError(
                "Failed to import cohere. Make sure you install cohere extra "
                "dependencies by running: "
                "pip install canopy-sdk[cohere]"
            )

        try:
            self._client = cohere.Client(api_key, api_url=base_url)
        except cohere.error.CohereError as e:
            raise RuntimeError(
                "Failed to connect to Cohere, please make sure that the CO_API_KEY "
                "environment variable is set correctly.\n"
                f"Error: {e.message}"
            )

        self.ignore_unrecognized_params = ignore_unrecognized_params
        self.default_model_params = kwargs

    def chat_completion(self,
                        system_prompt: str,
                        chat_history: Messages,
                        context: Optional[Context] = None,
                        *,
                        stream: bool = False,
                        max_tokens: Optional[int] = None,
                        model_params: Optional[dict] = None,
                        ) -> Union[ChatResponse, Iterable[StreamingChatChunk]]:
        """
        Chat completion using the Cohere API.

        Note: this function is wrapped in a retry decorator to handle transient errors.

        Args:
            system_prompt: The system prompt to use for the chat completion (preamble).
            chat_history: Messages (chat history) to send to the model.
            context: Knowledge base context to use for the chat completion. Defaults to None (no context).
            stream: Whether to stream the response or not.
            max_tokens: Maximum number of tokens to generate. Defaults to None (generates until stop sequence or until hitting max context size).
            model_params: Model parameters to use for this request. Defaults to None (uses the default model parameters).
                          Dictonary of parametrs to override the default model parameters if set on initialization.
                          For example, you can pass: {"temperature": 0.9, "top_p": 1.0} to override the default temperature and top_p.
                          see: https://platform.openai.com/docs/api-reference/chat/create
        Returns:
            ChatResponse or StreamingChatChunk

        Usage:
            >>> from canopy.llm import OpenAILLM
            >>> from canopy.models.data_models import UserMessage
            >>> llm = CohereLLM()
            >>> messages = [UserMessage(content="Hello! How are you?")]
            >>> result = llm.chat_completion(messages)
            >>> print(result.choices[0].message.content)
            "I'm good, how are you?"
        """  # noqa: E501
        model_params_dict: Dict[str, Any] = deepcopy(self.default_model_params)
        model_params_dict.update(
            model_params or {}
        )
        model_params_dict["max_tokens"] = max_tokens

        model_params_dict = self._convert_model_params(model_params_dict)

        connectors = model_params_dict.pop('connectors', None)
        messages: List[Dict[str, Any]] = self._map_messages(chat_history)
        model_name = model_params_dict.pop('model', None) or self.model_name

        if not messages:
            raise RuntimeError("No message provided")

        if system_prompt:
            messages = self._prepend_system_prompt_to_messages(system_prompt, messages)

        try:
            response = self._client.chat(
                model=model_name,
                message=messages.pop()['message'],
                chat_history=messages,
                documents=self.generate_documents_from_context(context),
                preamble_override=None,
                stream=stream,
                connectors=[
                    {"id": connector} for connector in connectors
                ] if connectors else None,
                **model_params_dict
            )
        except cohere.error.CohereAPIError as e:
            raise RuntimeError(
                f"Failed to use Cohere's {model_name} model for chat "
                f"completion. "
                f"Underlying Error:\n{e.message}"
            )

        def streaming_iterator(res):
            for chunk in res:
                if chunk.event_type != "text-generation":
                    continue

                choice = _StreamChoice(
                    index=0,
                    delta={
                        "content": chunk.text,
                        "function_call": None,
                        "role": Role.ASSISTANT,
                        "tool_calls": None
                    },
                    finish_reason=None,
                )
                streaming_chat_chunk = StreamingChatChunk(
                    id='',
                    object="chat.completion.chunk",
                    created=int(time.time()),
                    model=self.model_name,
                    choices=[choice],
                )
                streaming_chat_chunk.id = chunk.id

                yield streaming_chat_chunk

        if stream:
            return streaming_iterator(response)

        return ChatResponse(
            id=response.id,
            created=int(time.time()),
            choices=[_Choice(
                index=0,
                message=MessageBase(
                    role=Role.ASSISTANT,
                    content=response.text,
                ),
                finish_reason="stop",
            )],
            object="chat.completion",
            model=self.model_name,
            usage=TokenCounts(
                prompt_tokens=response.token_count["prompt_tokens"],
                completion_tokens=response.token_count["response_tokens"],
                total_tokens=response.token_count["billed_tokens"],
            ),
        )

    @retry(
        reraise=True,
        stop=stop_after_attempt(3),
    )
    def generate_search_queries(self, messages):
        messages = self._map_messages(messages)
        response = self._client.chat(
            model=self.model_name,
            message=messages[-1]['message'],
            chat_history=messages[:-1],
            stream=False,
            search_queries_only=True,
        )
        return [search_query['text'] for search_query in response.search_queries]

    def enforced_function_call(self,
                               system_prompt: str,
                               chat_history: Messages,
                               function: Function,
                               *,
                               max_tokens: Optional[int] = None,
                               model_params: Optional[dict] = None
                               ) -> dict:
        raise NotImplementedError("Cohere LLM doesn't support function calling")

    async def aenforced_function_call(self,
                                      system_prompt: str,
                                      chat_history: Messages,
                                      function: Function, *,
                                      max_tokens: Optional[int] = None,
                                      model_params: Optional[dict] = None):
        raise NotImplementedError("Cohere LLM doesn't support function calling")

    async def achat_completion(self,
                               system_prompt: str,
                               chat_history: Messages,
                               context: Optional[Context] = None,
                               *,
                               stream: bool = False,
                               max_generated_tokens: Optional[int] = None,
                               model_params: Optional[dict] = None,
                               ) -> Union[ChatResponse,
                                          Iterable[StreamingChatChunk]]:
        raise NotImplementedError("Cohere LLM doesn't support async chat completion")

    async def agenerate_queries(self,
                                messages: Messages,
                                *,
                                max_generated_tokens: Optional[int] = None,
                                model_params: Optional[dict] = None,
                                ) -> List[Query]:
        raise NotImplementedError("Cohere LLM doesn't support async query generation")

    def _convert_model_params(self, openai_model_params: dict) -> dict:
        """
        Convert Open AI model params to Cohere equivalents.

        Args:
            openai_model_params: model params passed from client to Canopy API in OpenAI format.

        Returns:
            Model params used with Cohere Chat API.
        """  # noqa: E501
        converted_model_params = {}

        for param in list(openai_model_params.keys()):
            if param in COMMON_PARAMS:
                converted_model_params[param] = openai_model_params.pop(param)
            elif param in EQUIVALENT_PARAMS:
                converted_model_params[EQUIVALENT_PARAMS[param]] = \
                    openai_model_params.pop(param)

        # Scale is -2.0 to 2.0 with OpenAI, but -1.0 to 1.0 with Cohere.
        if presence_penalty := converted_model_params.get("presence_penalty"):
            converted_model_params = presence_penalty * 0.5

        unrecognized_keys = set(openai_model_params.keys())
        default_keys = set(self.default_model_params.keys())

        if unrecognized_keys.difference(default_keys) \
                and not self.ignore_unrecognized_params:
            raise NotImplementedError(
                f"{','.join(unrecognized_keys)} not supported by Cohere Chat API."
            )

        return converted_model_params

    def _map_messages(self, messages: Messages) -> List[dict[str, Any]]:
        """
        Map the messages to format expected by Cohere.

        Cohere Chat API expects message history to be in the format:
        {
          "role": "USER",
          "message": "message text"
        }

        System messages will be passed as user messages.

        Args:
            messages: (chat history) to send to the model.

        Returns:
            list A List of dicts in format expected by Cohere chat API.
        """
        mapped_messages = []

        for message in messages:
            if not message.content:
                continue

            mapped_messages.append({
                "role": "CHATBOT" if message.role == Role.ASSISTANT else "USER",
                "message": message.content,
            })

        return mapped_messages

    def _prepend_system_prompt_to_messages(self,
                                           system_prompt: str,
                                           messages: List[dict[str, Any]]) -> (
                                                List)[dict[str, Any]]:
        """
        Prepend the value passed as the system prompt to the messages.

        Cohere does not have a direct equivalent to the system prompt, and when passing
        documents it's preferred to send the system prompt as the first message instead.
        """
        first_message = messages[0]

        if (first_message["message"] == system_prompt
                and first_message["role"] == "USER"):
            return messages

        system_prompt_messages = [
            {
                "role": "USER",
                "message": system_prompt,
            },
            {
                "role": "CHATBOT",
                "message": "Ok."
            }
        ]

        return system_prompt_messages + messages

    def generate_documents_from_context(
            self, context: Optional[Context]) -> List[Dict[str, Any]]:
        """
        Generate document data to pass to Cohere Chat API from provided context data.

        Args:
            context: Knowledge base context to use for the chat completion.

        Returns:
            documents: list of document objects for Cohere API.
        """
        if not context:
            return []

        if isinstance(context.content, StuffingContextContent):
            return (
                self.generate_documents_from_stuffing_context_content(context.content)
            )

        raise NotImplementedError(
            "Cohere LLM is currently supported only with StuffingContextBuilder."
        )

    def generate_documents_from_stuffing_context_content(
            self,
            content: StuffingContextContent) -> List[Dict[str, Any]]:
        """
        Generate document data to pass to Cohere Chat API from StuffingContextContent.

        Args:
            content: Stuffing context content from the context.

        Returns:
            documents: list of document objects for Cohere API.
        """
        documents = []

        for result in content.__root__:
            for snippet in result.snippets:
                documents.append(snippet.dict())

        return documents



from typing import Optional
from pinecone_text.dense import SentenceTransformerEncoder
from canopy.knowledge_base.record_encoder.dense import DenseRecordEncoder
from huggingface_hub.utils import RepositoryNotFoundError


class SentenceTransformerRecordEncoder(DenseRecordEncoder):
    """
    SentenceTransformerRecordEncoder is a type of DenseRecordEncoder that uses a Sentence Transformer model.
    The implementation uses the `SentenceTransformerEncoder` class from the `pinecone-text` library.
    For more information about see: https://github.com/pinecone-io/pinecone-text

    """  # noqa: E501

    def __init__(self,
                 *,
                 model_name: str = "sentence-transformers/all-MiniLM-L6-v2",
                 query_encoder_name: Optional[str] = None,
                 batch_size: int = 400,
                 device: Optional[str] = None,
                 **kwargs) -> None:
        """
        Initialize the SentenceTransformerRecordEncoder

        Args:
            model_name: The name of the embedding model to use for encoding documents.
                        See https://huggingface.co/models?library=sentence-transformers
                        for all possible Sentence Transformer models.
            query_encoder_name: The name of the embedding model to use for encoding queries.
                        See https://huggingface.co/models?library=sentence-transformers
                        for all possible Sentence Transformer models.
                        Defaults to `model_name`.
            batch_size: The number of documents or queries to encode at once.
                        Defaults to 400.
            device: The local device to use for encoding, for example "cpu", "cuda" or "mps".
                        Defaults to "cuda" if cuda is available, otherwise to "cpu".
            **kwargs: Additional arguments to pass to the underlying `pinecone-text.SentenceTransformerEncoder`.
        """  # noqa: E501
        try:
            encoder = SentenceTransformerEncoder(
                document_encoder_name=model_name,
                query_encoder_name=query_encoder_name,
                device=device,
                **kwargs,
            )
        except RepositoryNotFoundError as e:
            raise RuntimeError(
                "Your chosen Sentence Transformer model(s) could not be found. "
                f"Details: {str(e)}"
            ) from e
        except ImportError:
            raise ImportError(
                f"{self.__class__.__name__} requires the `torch` and `transformers` "
                f"extra dependencies. Please install them using "
                f"`pip install canopy-sdk[torch,transformers]`."
            )
        super().__init__(dense_encoder=encoder, batch_size=batch_size)



import os
from typing import List, Optional


from canopy.knowledge_base.models import KBQueryResult
from canopy.knowledge_base.reranker import Reranker

try:
    import cohere
    from cohere import CohereAPIError
except (OSError, ImportError, ModuleNotFoundError):
    _cohere_installed = False
else:
    _cohere_installed = True


class CohereReranker(Reranker):
    """
    Reranker that uses Cohere's text embedding to rerank documents.

    For each query and documents returned for that query, returns a list
    of documents ordered by their relevance to the provided query.
    """

    def __init__(self,
                 model_name: str = 'rerank-english-v2.0',
                 *,
                 top_n: int = 10,
                 api_key: Optional[str] = None):
        """
            Initializes the Cohere reranker.

            Args:
                model_name: The identifier of the model to use, one of :
                    rerank-english-v2.0, rerank-multilingual-v2.0
                top_n: The number of most relevant documents return, defaults to 10
                api_key: API key for Cohere. If not passed `CO_API_KEY` environment
                    variable will be used.
        """

        if not _cohere_installed:
            raise ImportError(
                "Failed to import cohere. Make sure you install cohere extra "
                "dependencies by running: "
                "pip install canopy-sdk[cohere]"
            )
        cohere_api_key = api_key or os.environ.get("CO_API_KEY")
        if cohere_api_key is None:
            raise RuntimeError(
                "Cohere API key is required to use Cohere Reranker. "
                "Please provide it as an argument "
                "or set the CO_API_KEY environment variable."
            )
        self._client = cohere.Client(api_key=cohere_api_key)
        self._model_name = model_name
        self._top_n = top_n

    def rerank(self, results: List[KBQueryResult]) -> List[KBQueryResult]:
        reranked_query_results: List[KBQueryResult] = []
        for result in results:
            texts = [doc.text for doc in result.documents]
            try:
                response = self._client.rerank(query=result.query,
                                               documents=texts,
                                               top_n=self._top_n,
                                               model=self._model_name)
            except CohereAPIError as e:
                raise RuntimeError("Failed to rerank documents using Cohere."
                                   f" Underlying Error:\n{e.message}")

            reranked_docs = []
            for rerank_result in response:
                doc = result.documents[rerank_result.index].copy(
                    deep=True,
                    update=dict(score=rerank_result.relevance_score)
                )
                reranked_docs.append(doc)

            reranked_query_results.append(KBQueryResult(query=result.query,
                                                        documents=reranked_docs))
        return reranked_query_results

    async def arerank(self, results: List[KBQueryResult]) -> List[KBQueryResult]:
        raise NotImplementedError()



import logging
import re
from typing import List, Optional, cast

from tenacity import retry, stop_after_attempt, retry_if_exception_type

from canopy.chat_engine.history_pruner.raising import RaisingHistoryPruner
from canopy.chat_engine.query_generator import QueryGenerator, LastMessageQueryGenerator
from canopy.llm import BaseLLM, OpenAILLM
from canopy.models.api_models import ChatResponse
from canopy.models.data_models import Messages, Query, UserMessage

logger = logging.getLogger(__name__)

SYSTEM_PROMPT = """You are an expert on formulating a search query for a search engine,
to assist in responding to the user's question.

Given the following conversation, create a standalone question summarizing
 the user's last question, in its original language.

Reply to me in JSON in this format:

 {"question": {The question you generated here}}.

   Example:

        User: What is the weather today?

        Expected Response:
            ```json
            {"question": "What is the weather today?"}
            ```

    Example 2:

        User: How should I wash my white clothes in the laundry?
        Assistant: Separate from the colorful ones, and use a bleach.
        User: Which temperature?

        Expected Response:
         ```json
         {"question": "What is the right temperature for washing white clothes?"}
         ```


   Do not try to answer the question; just try to formulate a question representing the user's question.
   Do not return any other format other than the specified JSON format and keep it really short.

"""  # noqa: E501

USER_PROMPT = "Return only a JSON containing a single key 'question' and the value."


class ExtractionException(ValueError):
    pass


class InstructionQueryGenerator(QueryGenerator):
    _DEFAULT_COMPONENTS = {
        "llm": OpenAILLM,
    }

    def __init__(self,
                 *,
                 llm: Optional[BaseLLM] = None):
        """
             This `QueryGenerator` uses an LLM to formulate a knowledge base query
             from the full chat history. It does so by prompting the LLM to reply
             with a JSON containing a single key `question`, containing the query
             for the knowledge base. If LLM response cannot be parsed
             (after multiple retries), it falls back to returning the last message
             from the history as a query, much like `LastMessageQueryGenerator`
        """
        self._llm = llm or self._DEFAULT_COMPONENTS["llm"]()
        self._system_prompt = SYSTEM_PROMPT
        self._history_pruner = RaisingHistoryPruner()
        self._last_message_query_generator = LastMessageQueryGenerator()

        # Define a regex pattern to find the JSON object with the key "question"
        self._question_regex = re.compile(r'{\s*"question":\s*"([^"]+)"\s*}')

    def generate(self,
                 messages: Messages,
                 max_prompt_tokens: int) -> List[Query]:

        # Add a user message at the end; that helps us return a JSON object.
        new_history = (
                messages +
                [
                    UserMessage(content=USER_PROMPT)
                ]
        )

        chat_history = self._history_pruner.build(system_prompt=self._system_prompt,
                                                  chat_history=new_history,
                                                  max_tokens=max_prompt_tokens)

        question = self._try_generate_question(chat_history)

        if question is None:
            logger.warning("Falling back to the last message query generator.")
            return self._last_message_query_generator.generate(messages, 0)
        else:
            return [Query(text=question)]

    def _get_answer(self, messages: Messages) -> str:
        llm_response = self._llm.chat_completion(system_prompt=self._system_prompt,
                                                 chat_history=messages)
        response = cast(ChatResponse, llm_response)
        return response.choices[0].message.content

    @retry(stop=stop_after_attempt(3),
           retry=retry_if_exception_type(ExtractionException),
           retry_error_callback=lambda _: None)
    def _try_generate_question(self, messages: Messages) -> Optional[str]:
        content = self._get_answer(messages)
        return self._extract_question(content)

    def _extract_question(self, text: str) -> str:

        # Search for the pattern in the text
        match = re.search(self._question_regex, text)

        # If a match is found, extract and return the first occurrence
        if match:
            return match.group(1)

        raise ExtractionException("Failed to extract the question.")

    async def agenerate(self,
                        messages: Messages,
                        max_prompt_tokens: int) -> List[Query]:
        raise NotImplementedError



import os
from typing import Optional, Any

import openai

from canopy.llm import OpenAILLM


class AzureOpenAILLM(OpenAILLM):
    """
    Azure OpenAI LLM wrapper built on top of the OpenAI Python client.

    Azure OpenAI services require a valid API key, and an Azure endpoint. You will need
    To set the following environment variables:
    - AZURE_OPENAI_API_KEY: Your Azure OpenAI API key.
    - AZURE_OPENAI_ENDPOINT: Your Azure endpoint, including the resource, e.g. `https://example-resource.azure.openai.com/`

    To use a specific custom model deployment, simply pass the deployment name to the `model_name` argument

    Note: If you want to set an OpenAI organization, you would need to set environment variable "OPENAI_ORG_ID".
          This is different from the environment variable for passing an organization to the parent class (OpenAILLM), which is "OPENAI_ORG".
    """  # noqa: E501

    def __init__(self,
                 model_name: str,
                 *,
                 api_key: Optional[str] = None,
                 api_version: str = "2023-12-01-preview",
                 azure_endpoint: Optional[str] = None,
                 **kwargs: Any,
                 ):
        """
        Initialize the Azure OpenAI LLM.

        Args:
            model_name: The name of your custom model deployment on Azure. This is required.
            api_key: Your Azure OpenAI API key. Defaults to None (uses the "AZURE_OPENAI_API_KEY" environment variable).
            api_version: The Azure OpenAI API version to use. Defaults to "2023-12-01-preview".
            azure_endpoint: The url of your Azure OpenAI service endpoint. Defaults to None (uses the "AZURE_OPENAI_ENDPOINT" environment variable).
            **kwargs: Generation default parameters to use for each request.


        >>> from canopy.llm.azure_openai_llm import AzureOpenAILLM
        >>> from canopy.models.data_models import UserMessage
        >>> llm = AzureOpenAILLM()
        >>> messages = [UserMessage(content="Hello! How are you?")]
        >>> llm.chat_completion(messages)

        """  # noqa: E501
        self.model_name = model_name

        try:
            self._client = openai.AzureOpenAI(
                azure_deployment=model_name,
                api_key=api_key,
                api_version=api_version,
                azure_endpoint=azure_endpoint,  # type: ignore
            )
        except (openai.OpenAIError, ValueError) as e:
            raise RuntimeError(
                "Failed to connect to Azure OpenAI, please make sure that the "
                "AZURE_OPENAI_API_KEY and AZURE_OPENAI_ENDPOINT environment variables "
                "are set correctly. "
                f"Underlying Error:\n{self._format_openai_error(e)}"
            ) from e

        self.default_model_params = kwargs

    @property
    def available_models(self):
        raise NotImplementedError(
            "Azure OpenAI LLM does not support listing available models"
        )

    def _handle_chat_error(self, e, is_function_call=False):
        if isinstance(e, openai.AuthenticationError):
            raise RuntimeError(
                "Failed to connect to Azure OpenAI, please make sure that the "
                "AZURE_OPENAI_API_KEY environment variable is set correctly. "
                f"Underlying Error:\n{self._format_openai_error(e)}"
            ) from e
        elif isinstance(e, openai.APIConnectionError):
            raise RuntimeError(
                f"Failed to connect to your Azure OpenAI endpoint, please make sure "
                f"that the provided endpoint {os.getenv('AZURE_OPENAI_ENDPOINT')} "
                f"is correct. Underlying Error:\n{self._format_openai_error(e)}"
            ) from e
        elif isinstance(e, openai.NotFoundError):
            if e.type and 'invalid' in e.type and is_function_call:
                raise NotImplementedError(
                    f"It seems that you are trying to use OpenAI's `function calling` "
                    f"or `tools` features. Please note that Azure OpenAI only supports "
                    f"function calling for specific models and API versions. More "
                    f"details in: https://learn.microsoft.com/en-us/azure/ai-services/openai/how-to/function-calling. "  # noqa: E501
                    f"Underlying Error:\n{self._format_openai_error(e)}"
                ) from e
            else:
                raise RuntimeError(
                    f"Failed to connect to your Azure OpenAI. Please make sure that "
                    f"you have provided the correct deployment name: {self.model_name} "
                    f"and API version: {self._client._api_version}. "
                    f"Underlying Error:\n{self._format_openai_error(e)}"
                ) from e
        else:
            super()._handle_chat_error(e)



import os

from pinecone_text.dense import AzureOpenAIEncoder

from canopy.knowledge_base.record_encoder import OpenAIRecordEncoder, DenseRecordEncoder
import openai


class AzureOpenAIRecordEncoder(OpenAIRecordEncoder):
    """
    AzureOpenAIRecordEncoder is a type of DenseRecordEncoder that uses the Azure OpenAI's `embeddings` deployments.
    The implementation uses the `AzureOpenAIEncoder` class from the `pinecone-text` library.
    For more information about see: https://github.com/pinecone-io/pinecone-text

    Azure OpenAI services require a valid API key, and an Azure endpoint. You will need
    To set the following environment variables:
    - AZURE_OPENAI_API_KEY: Your Azure OpenAI API key.
    - AZURE_OPENAI_ENDPOINT: Your Azure endpoint, including the resource, e.g. `https://example-resource.azure.openai.com/`
    """  # noqa: E501

    def __init__(
            self,
            *,
            model_name: str,
            api_version: str = "2023-12-01-preview",
            batch_size: int = 400,
            **kwargs
    ):
        """
        Initialize the AzureOpenAIRecordEncoder

        Args:
            model_name: The name of embeddings model deployment to use for encoding
            api_version: The Azure OpenAI API version to use. Defaults to "2023-12-01-preview".
            batch_size: The number of documents or queries to encode at once.
                        Defaults to 400.
            **kwargs: Additional arguments to pass to the underlying `pinecone-text.AzureOpenAIEncoder`.
        """  # noqa: E501
        try:
            encoder = AzureOpenAIEncoder(model_name, api_version=api_version, **kwargs)
        except (openai.OpenAIError, ValueError) as e:
            raise RuntimeError(
                "Failed to connect to Azure OpenAI, please make sure that the "
                "AZURE_OPENAI_API_KEY and AZURE_OPENAI_ENDPOINT environment variables "
                "are set correctly. "
                f"Underlying Error:\n{self._format_openai_error(e)}"
            ) from e

        DenseRecordEncoder.__init__(self, dense_encoder=encoder, batch_size=batch_size,
                                    **kwargs)

    def _format_error(self, err):
        if isinstance(err, openai.AuthenticationError):
            return (
                "Failed to connect to Azure OpenAI, please make sure that the "
                "AZURE_OPENAI_API_KEY environment variable is set correctly. "
                f"Underlying Error:\n{self._format_openai_error(err)}"
            )
        elif isinstance(err, openai.APIConnectionError):
            return (
                f"Failed to connect to your Azure OpenAI endpoint, please make sure "
                f"that the provided endpoint {os.getenv('AZURE_OPENAI_ENDPOINT')} "
                f"is correct. Underlying Error:\n{self._format_openai_error(err)}"
            )
        elif isinstance(err, openai.NotFoundError):
            return (
                f"Failed to connect to your Azure OpenAI. Please make sure that "
                f"you have provided the correct deployment name: {self.model_name} "
                f"and API version: {self._client._api_version}. "
                f"Underlying Error:\n{self._format_openai_error(err)}"
            )
        else:
            return super()._format_error(err)



from typing import List, Optional

from tokenizers import Tokenizer as HfTokenizer
try:
    import cohere
except (OSError, ImportError, ModuleNotFoundError):
    _cohere_installed = False
else:
    _cohere_installed = True

from .base import BaseTokenizer
from ..models.data_models import Messages


class CohereHFTokenizer(BaseTokenizer):
    """
    Tokenizer for Cohere models, based on the Hugging Face tokenizers library.

    Usage:
    Initialize the singleton tokenizer with the LlamaTokenizer class:
    >>> from canopy.tokenizer import Tokenizer
    >>> Tokenizer.initialize(tokenizer_class=CohereHFTokenizer,
                             model_name="Cohere/Command-nightly")

    You can then use the tokenizer instance from anywhere in the code:
    >>> from canopy.tokenizer import Tokenizer
    >>> tokenizer = Tokenizer()
    >>> tokenizer.tokenize("Hello World!")
    ['▁Hello', '▁World', '!']
    """

    MESSAGE_TOKENS_OVERHEAD = 3
    FIXED_PREFIX_TOKENS = 3

    def __init__(
        self,
        model_name: str = "Cohere/Command-nightly",
    ):
        """
        Initialize the tokenizer.

        Args:
            model_name: The name of the Hugging Face model to use. Defaults to "Cohere/Command-nightly".
        """  # noqa: E501
        if not _cohere_installed:
            raise ImportError(
                "Failed to import cohere. Make sure you install cohere extra "
                "dependencies by running: "
                "pip install canopy-sdk[cohere]"
            )

        self._encoder = HfTokenizer.from_pretrained(model_name)

    def tokenize(self, text: str) -> List[str]:
        """
        Tokenize a text using HuggingFace's tokenizers library.

        Args:
            text: The text to tokenize.

        Returns:
            The list of tokens.
        """
        return self._encoder.encode(text, add_special_tokens=False).tokens

    def detokenize(self, tokens: List[str]) -> str:
        """
        Detokenize a list of tokens that were previously tokenized using this tokenizer.

        Args:
            tokens: The list of tokens to detokenize.

        Returns:
            The detokenized text as a string.
        """
        if not isinstance(tokens, List):
            raise TypeError(f"detokenize expect List[str], got f{type(tokens)}")

        ids = [self._encoder.token_to_id(token) for token in tokens]
        return self._encoder.decode(ids)

    def token_count(self, text: str) -> int:
        """
        Count the number of tokens in a text.

        Args:
            text: The text to count the tokens of.

        Returns:
            The number of tokens in the text.
        """
        return len(self._encoder.encode(text, add_special_tokens=False).ids)

    def messages_token_count(self, messages: Messages) -> int:
        """
        Count the number of tokens in a list of messages, as expected to be
        counted by Cohere models.

        Args:
            messages: The list of messages to count the tokens of.

        Returns:
            The number of tokens in the messages, as expected to be counted by Cohere models.
        """  # noqa: E501
        num_tokens = 0
        for message in messages:
            num_tokens += self.MESSAGE_TOKENS_OVERHEAD
            for key, value in message.dict().items():
                num_tokens += self.token_count(value)
        num_tokens += self.FIXED_PREFIX_TOKENS
        return num_tokens


class CohereAPITokenizer(BaseTokenizer):
    """
    Tokenizer for Cohere models, based on the Cohere Tokenize API.

    Usage:
    Initialize the singleton tokenizer with the CohereAPITokenizer class:
    >>> from canopy.tokenizer import Tokenizer
    >>> Tokenizer.initialize(tokenizer_class=CohereAPITokenizer, model_name="embed-multilingual-v3.0")

    You can then use the tokenizer instance from anywhere in the code:
    >>> from canopy.tokenizer import Tokenizer
    >>> tokenizer = Tokenizer()
    >>> tokenizer.tokenize("Hello world!")
    ['Hello', ' world', '!']
    """  # noqa: E501

    MESSAGE_TOKENS_OVERHEAD = 3
    FIXED_PREFIX_TOKENS = 3

    def __init__(self,
                 model_name: Optional[str] = None,
                 *,
                 api_key: Optional[str] = None,
                 api_url: Optional[str] = None):
        """
        Initialize the tokenizer.

        Args:
            model_name: The name of the model to use.
            api_key: Your Cohere API key. Defaults to None (uses the "CO_API_KEY" environment variable).
            api_url: The base URL to use for the Cohere API. Defaults to None (uses the "CO_API_URL" environment variable if set, otherwise use default Cohere API URL).
        """  # noqa: E501
        self.model_name = model_name
        self._client = cohere.Client(api_key, api_url=api_url)

    def tokenize(self, text: str) -> List[str]:
        """
        Tokenize a text using Cohere Tokenize API.

        Args:
            text: The text to tokenize.

        Returns:
            The list of tokens.
        """
        if not text:
            return []

        tokens = self._client.tokenize(text, model=self.model_name)
        return tokens.token_strings

    def detokenize(self, tokens: List[str]) -> str:
        """
        Detokenize a list of tokens that were previously tokenized using this tokenizer.

        Args:
            tokens: The list of tokens to detokenize.

        Returns:
            The detokenized text as a string.
        """
        if not isinstance(tokens, List):
            raise TypeError(f"detokenize expects List[str], got f{type(tokens)}")
        return "".join(tokens)

    def messages_token_count(self, messages: Messages) -> int:
        """
        Count the number of tokens in a list of messages as expected to be counted by Cohere models.
        Account for the overhead of the messages structure.
        Taken from: https://github.com/openai/openai-cookbook/.../How_to_format_inputs_to_ChatGPT_models.ipynb

        Args:
            messages: The list of messages to count the tokens of.

        Returns:
            The number of tokens in the messages, as expected to be counted by OpenAI models.
        """  # noqa: E501
        num_tokens = 0
        for message in messages:
            num_tokens += self.MESSAGE_TOKENS_OVERHEAD
            for key, value in message.dict().items():
                num_tokens += self.token_count(value)
        num_tokens += self.FIXED_PREFIX_TOKENS
        return num_tokens



from typing import List

from canopy.knowledge_base.models import KBQueryResult
from canopy.knowledge_base.reranker import Reranker


class TransparentReranker(Reranker):
    """
    Transparent reranker that does nothing, it just returns the results as is. This is the default reranker.
    The TransparentReranker is used as a placeholder for future development "forcing" every result set to be reranked.
    """  # noqa: E501

    def rerank(self, results: List[KBQueryResult]) -> List[KBQueryResult]:
        """
        Returns the results as is.

        Args:
            results: A list of KBQueryResult to rerank.

        Returns:
            results: A list of KBQueryResult, same as the input.
        """  # noqa: E501
        return results

    async def arerank(self, results: List[KBQueryResult]) -> List[KBQueryResult]:
        return results



from pinecone_text.dense import JinaEncoder
from canopy.knowledge_base.record_encoder.dense import DenseRecordEncoder


class JinaRecordEncoder(DenseRecordEncoder):
    """
    JinaRecordEncoder is a type of DenseRecordEncoder that uses the JinaAI `embeddings` API.
    The implementation uses the `JinaEncoder` class from the `pinecone-text` library.
    For more information about see: https://github.com/pinecone-io/pinecone-text

    """  # noqa: E501

    def __init__(self,
                 *,
                 model_name: str = "jina-embeddings-v2-base-en",
                 batch_size: int = 400,
                 **kwargs):
        """
        Initialize the JinaRecordEncoder

        Args:
            model_name: The name of the embedding model to use.
            batch_size: The number of documents or queries to encode at once.
                        Defaults to 400.
            **kwargs: Additional arguments to pass to the underlying `pinecone-text. JinaEncoder`.
        """  # noqa: E501
        encoder = JinaEncoder(model_name, **kwargs)
        super().__init__(dense_encoder=encoder, batch_size=batch_size)



from abc import ABC, abstractmethod
from typing import Optional

from canopy.tokenizer import Tokenizer
from canopy.models.data_models import Messages, SystemMessage, Context
from canopy.utils.config import ConfigurableMixin


class HistoryPruner(ABC, ConfigurableMixin):

    def __init__(self):
        self._tokenizer = Tokenizer()

    @abstractmethod
    def build(self,
              chat_history: Messages,
              max_tokens: int,
              system_prompt: Optional[str] = None,
              context: Optional[Context] = None,
              ) -> Messages:
        raise NotImplementedError

    async def abuild(self,
                     chat_history: Messages,
                     max_tokens: int) -> Messages:
        raise NotImplementedError()

    def _max_tokens_history(self,
                            max_tokens: int,
                            system_prompt: Optional[str] = None,
                            context: Optional[Context] = None, ) -> int:
        if system_prompt is not None:
            max_tokens -= self._tokenizer.messages_token_count(
                [SystemMessage(content=system_prompt)]
            )

        if context is not None:
            max_tokens -= context.num_tokens

        return max_tokens



from typing import Optional, Any
import os
from canopy.llm import OpenAILLM
from canopy.llm.models import Function
from canopy.models.data_models import Messages

ANYSCALE_BASE_URL = "https://api.endpoints.anyscale.com/v1"
FUNCTION_MODEL_LIST = [
    "mistralai/Mistral-7B-Instruct-v0.1",
    "mistralai/Mixtral-8x7B-Instruct-v0.1",
]


class AnyscaleLLM(OpenAILLM):
    """
    Anyscale LLM wrapper built on top of the OpenAI Python client.

    Note: Anyscale requires a valid API key to use this class.
          You can set the "ANYSCALE_API_KEY" environment variable.
    """

    def __init__(
        self,
        model_name: str = "meta-llama/Llama-2-7b-chat-hf",
        *,
        base_url: Optional[str] = ANYSCALE_BASE_URL,
        api_key: Optional[str] = None,
        **kwargs: Any,
    ):
        ae_api_key = api_key or os.environ.get("ANYSCALE_API_KEY")
        if not ae_api_key:
            raise ValueError(
                "Anyscale API key is required to use Anyscale. "
                "Please provide it as an argument "
                "or set the ANYSCALE_API_KEY environment variable."
            )
        ae_base_url = base_url
        super().__init__(model_name, api_key=ae_api_key, base_url=ae_base_url, **kwargs)

    def enforced_function_call(
        self,
        system_prompt: str,
        chat_history: Messages,
        function: Function,
        *,
        max_tokens: Optional[int] = None,
        model_params: Optional[dict] = None,
    ) -> dict:
        model = self.model_name
        if model_params and "model" in model_params:
            model = model_params["model"]
        if model not in FUNCTION_MODEL_LIST:
            raise NotImplementedError(
                f"Model {model} doesn't support function calling. "
                "To use function calling capability, please select a different model.\n"
                "Pleaes check following link for details: "
                "https://docs.endpoints.anyscale.com/guides/function-calling"
            )
        else:
            return super().enforced_function_call(
                system_prompt, chat_history, function,
                max_tokens=max_tokens, model_params=model_params
            )

    def aenforced_function_call(self,
                                system_prompt: str,
                                chat_history: Messages,
                                function: Function,
                                *,
                                max_tokens: Optional[int] = None,
                                model_params: Optional[dict] = None
                                ):
        raise NotImplementedError()



import logging
from functools import cached_property
from typing import List, Optional

from pinecone_text.hybrid import hybrid_convex_scale
from pinecone_text.sparse import BM25Encoder

from . import DenseRecordEncoder, OpenAIRecordEncoder
from .base import RecordEncoder
from canopy.knowledge_base.models import KBQuery, KBEncodedDocChunk, KBDocChunk
from canopy.models.data_models import Query

logger = logging.getLogger(__name__)


class HybridRecordEncoder(RecordEncoder):
    """
    HybridRecordEncoder is a subclass of RecordEncoder that generates sparse and dense vector representation of
    documents` chunks and textual queries.

    The dense representation generated by the `HybridRecordEncoder` is a list of floats in a given dimension.
    The sparse representation generated by the `HybridRecordEncoder` is a `SparseVector`.

    HybridRecordEncoder uses DenseRecordEncoder for dense encoding and BM25Encoder for sparse encoding.

    Alpha is a parameter that controls the weight of the dense vector in the hybrid representation.
    If alpha is 1, the query vector will be the dense vector. The default value of alpha is 0.5.

    For more information about the encoders see: https://github.com/pinecone-io/pinecone-text

    """  # noqa: E501

    _DEFAULT_COMPONENTS = {
        "dense_record_encoder": OpenAIRecordEncoder
    }

    def __init__(self,
                 dense_record_encoder: Optional[DenseRecordEncoder] = None,
                 alpha: float = 0.5,
                 bm_25_encoder_df_path: Optional[str] = None,
                 **kwargs):
        """
        Initialize the encoder.

        Args:
            dense_record_encoder: A DenseRecordEncoder to encode the text.
            alpha: The weight of the dense vector in the hybrid representation (between 0 and 1).
            bm_25_encoder_df_path: The path to the file that contains the document frequencies of the BM25Encoder.\
            You can create this file by fitting the BM25Encoder on a corpus of documents and calling `dump`\
            on the encoder.
            **kwargs: Additional arguments to pass to the RecordEncoder.
        """  # noqa: E501

        if alpha == 0:
            raise ValueError("Sparse only representation is not supported. "
                             "Alpha must be greater than 0.")

        if not 0 < alpha <= 1:
            raise ValueError("Alpha must be between 0 (excluded) and 1 (included)")

        super().__init__(**kwargs)

        if dense_record_encoder:
            if not isinstance(dense_record_encoder, DenseRecordEncoder):
                raise TypeError(
                    f"dense_encoder must be an instance of DenseRecordEncoder, "
                    f"not {type(dense_record_encoder)}"
                )
            self._dense_record_encoder = dense_record_encoder
        else:
            default_dense_encoder = self._DEFAULT_COMPONENTS["dense_record_encoder"]
            self._dense_record_encoder = default_dense_encoder()

        self._bm_25_encoder_df_path = bm_25_encoder_df_path
        self._alpha = alpha

    @cached_property
    def _sparse_encoder(self) -> BM25Encoder:
        logger.info("Loading the document frequencies for the BM25Encoder...")
        if self._bm_25_encoder_df_path is None:
            encoder = BM25Encoder.default()
        else:
            encoder = BM25Encoder().load(self._bm_25_encoder_df_path)
        logger.info("Finished loading the document frequencies for the BM25Encoder.")
        return encoder

    def _encode_documents_batch(self,
                                documents: List[KBDocChunk]
                                ) -> List[KBEncodedDocChunk]:
        """
        Encode a batch of documents, takes a list of KBDocChunk and returns a list of KBEncodedDocChunk.

        Args:
            documents: A list of KBDocChunk to encode.
        Returns:
            encoded chunks: A list of KBEncodedDocChunk,
            with the `values` containing the generated dense vector and
            `sparse_values` containing the generated sparse vector.
        """  # noqa: E501

        chunks = self._dense_record_encoder.encode_documents(documents)
        sparse_values = self._sparse_encoder.encode_documents(
            [d.text for d in documents]
        )
        for chunk, sv in zip(chunks, sparse_values):
            chunk.sparse_values = sv
        return chunks

    def _encode_queries_batch(self, queries: List[Query]) -> List[KBQuery]:
        """
        Encode a batch of queries, takes a list of Query and returns a list of KBQuery.
        Args:
            queries: A list of Query to encode.
        Returns:
            encoded queries: A list of KBQuery, with the `values` containing the generated dense vector with the weight
            alpha and `sparse_values` containing the generated sparse vector with the weight (1 - alpha).
        """  # noqa: E501

        dense_queries = self._dense_record_encoder.encode_queries(queries)
        sparse_values = self._sparse_encoder.encode_queries([q.text for q in queries])

        scaled_values = [
            hybrid_convex_scale(dq.values, sv, self._alpha) for dq, sv in
            zip(dense_queries, sparse_values)
        ]

        return [q.copy(update=dict(values=v, sparse_values=sv)) for q, (v, sv) in
                zip(dense_queries, scaled_values)]

    @property
    def dimension(self) -> int:
        return self._dense_record_encoder.dimension

    async def _aencode_documents_batch(self,
                                       documents: List[KBDocChunk]
                                       ) -> List[KBEncodedDocChunk]:
        raise NotImplementedError

    async def _aencode_queries_batch(self, queries: List[Query]) -> List[KBQuery]:
        raise NotImplementedError



from typing import Optional

from canopy.chat_engine.history_pruner.base import HistoryPruner
from canopy.models.data_models import Messages, Context


class RecentHistoryPruner(HistoryPruner):

    def __init__(self,
                 min_history_messages: int = 1):
        super().__init__()
        self._min_history_messages = min_history_messages

    def build(self,
              chat_history: Messages,
              max_tokens: int,
              system_prompt: Optional[str] = None,
              context: Optional[Context] = None,
              ) -> Messages:
        max_tokens = self._max_tokens_history(max_tokens,
                                              system_prompt,
                                              context)
        token_count = self._tokenizer.messages_token_count(chat_history)
        if token_count < max_tokens:
            return chat_history

        truncated_history = chat_history[-self._min_history_messages:]
        token_count = self._tokenizer.messages_token_count(truncated_history)
        if token_count > max_tokens:
            raise ValueError(f"The {self._min_history_messages} most recent messages in"
                             f" history require {token_count} tokens, which exceeds the"
                             f" calculated limit for history of {max_tokens}"
                             f" tokens out of total {max_tokens} allowed"
                             f" in context window.")

        for message in reversed(chat_history[:-self._min_history_messages]):
            token_count = self._tokenizer.messages_token_count(
                truncated_history + [message]
            )

            # If the message can fit into the remaining tokens, add it
            if token_count > max_tokens:
                break

            truncated_history.insert(0, message)

        return truncated_history

    async def abuild(self,
                     chat_history: Messages,
                     max_tokens: int) -> Messages:
        raise NotImplementedError()
